import os
import requests
import json
from Crypto.Random import get_random_bytes
from Crypto.Util.Padding import pad
import base64
import string
import secrets
from datetime import datetime
from tqdm import tqdm
import hashlib
from Crypto.Cipher import AES
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import serialization
import requests

class Neuropacs:
    def __init__(self, server_url, api_key, client="api"):
        """
        NeuroPACS constructor
        """
        self.server_url = server_url
        self.api_key = api_key
        self.client = client
        self.aes_key = self.__generate_aes_key()
        self.connection_id = ""
        self.aes_key = ""
        self.dataset_upload = False
        self.files_uploaded = 0

    def __generate_aes_key(self):
        """Generate an 16-byte AES key for AES-CTR encryption.

        :return: AES key encoded as a base64 string.
        """
        aes_key = get_random_bytes(16)
        aes_key_base64 = base64.b64encode(aes_key).decode('utf-8')
        return aes_key_base64

    def __oaep_encrypt(self, plaintext):
        """
        OAEP encrypt plaintext.

        :param str/JSON plaintext: Plaintext to be encrypted.

        :return: Base64 string OAEP encrypted ciphertext
        """

        try:
            plaintext = json.dumps(plaintext)
        except:
            if not isinstance(plaintext, str):
                raise Exception({"neuropacsError": "Plaintext must be a string or JSON!"})    

    
        # get public key of server
        PUBLIC_KEY = self.get_public_key()

        PUBLIC_KEY = PUBLIC_KEY.encode('utf-8')

        # Deserialize the public key from PEM format
        public_key = serialization.load_pem_public_key(PUBLIC_KEY)

        # Encrypt the plaintext using OAEP padding
        ciphertext = public_key.encrypt(
            plaintext.encode('utf-8'),
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )
        ciphertext_key_base64 = base64.b64encode(ciphertext).decode('utf-8')

        # Return the ciphertext as bytes
        return ciphertext_key_base64

    def __encrypt_aes_ctr(self, plaintext, format_in, format_out):
        """AES CTR encrypt plaintext

        :param JSON/str/bytes plaintext: Plaintext to be encrypted.
        :param str format_in: format of plaintext. Defaults to "string".
        :param str format_out: format of ciphertext. Defaults to "string".

        :return: Encrypted ciphertext in requested format_out.
        """        

        plaintext_bytes = ""

        try:
            if format_in == "string" and isinstance(plaintext, str):
                plaintext_bytes = plaintext.encode("utf-8")
            elif format_in == "bytes" and isinstance(plaintext,bytes):
                plaintext_bytes = plaintext
            elif format_in == "json":
                plaintext_json = json.dumps(plaintext)
                plaintext_bytes = plaintext_json.encode("utf-8")
            else:
                raise Exception({"neuropacsError": "Invalid plaintext format!"})
        except Exception as e:
            if(isinstance(e.args[0], dict) and 'neuropacsError' in e.args[0]):
                raise Exception(e.args[0]['neuropacsError'])
            else:   
                raise Exception("Invalid plaintext format!")

        try:
            aes_key_bytes = base64.b64decode(self.aes_key)

            padded_plaintext = pad(plaintext_bytes, AES.block_size)

            # generate IV
            iv = get_random_bytes(16)

            # Create an AES cipher object in CTR mode
            cipher = AES.new(aes_key_bytes, AES.MODE_CTR, initial_value=iv, nonce=b'')

            # Encrypt the plaintext
            ciphertext = cipher.encrypt(padded_plaintext)

            # Combine IV and ciphertext
            encrypted_data = iv + ciphertext

            encryped_message = ""

            if format_out == "string":
                encryped_message = base64.b64encode(encrypted_data).decode('utf-8')
            elif format_out == "bytes":
                encryped_message = encrypted_data

            return encryped_message

        except Exception as e:
            if(isinstance(e.args[0], dict) and 'neuropacsError' in e.args[0]):
                raise Exception(e.args[0]['neuropacsError']) 
            else:
                raise Exception("AES encryption failed!")   

    def __decrypt_aes_ctr(self, encrypted_data, format_out):
        """AES CTR decrypt ciphertext.

        :param str ciphertext: Ciphertext to be decrypted.
        :param * format_out: Format of plaintext. Default to "string".

        :return: Plaintext in requested format_out.
        """

        try:

            aes_key_bytes = base64.b64decode(self.aes_key)

            # Decode the base64 encoded encrypted data
            encrypted_data = base64.b64decode(encrypted_data)

            # Extract IV and ciphertext
            iv = encrypted_data[:16]

            ciphertext = encrypted_data[16:]

            # Create an AES cipher object in CTR mode
            cipher = AES.new(aes_key_bytes, AES.MODE_CTR, initial_value=iv, nonce=b'')

            # Decrypt the ciphertext and unpad the result
            decrypted = cipher.decrypt(ciphertext)

            decrypted_data = decrypted.decode("utf-8")

            if format_out == "json":
                decrypted_data = json.loads(decrypted_data)
            elif format_out == "string":
                pass

            return decrypted_data
        except Exception as e:
            if(isinstance(e.args[0], dict) and 'neuropacsError' in e.args[0]):
                raise Exception(e.args[0]['neuropacsError']) 
            else:
                raise Exception("AES decryption failed!")
    
    def __generate_filename(self):
        """Generate a filename for byte data
        :return: 20 character random alphanumeric string
        """
        characters = string.ascii_letters + string.digits
        random_string = ''.join(secrets.choice(characters) for _ in range(20))
        return random_string

    def get_public_key(self, server_url=None):
        """Retrieve public key from server.

        :param str server_url: Server URL of Neuropacs instance

        :return: Base64 string public key.
        """

        if server_url is None:
            server_url = self.server_url

        try:
            res = requests.get(f"{server_url}/api/getPubKey")

            if not res.ok:
                raise Exception({"neuropacsError": f"{res.text}"})

            json = res.json()
            pub_key = json['pub_key']
            return pub_key
        except Exception as e:
            if(isinstance(e.args[0], dict) and 'neuropacsError' in e.args[0]):
                raise Exception(e.args[0]['neuropacsError']) 
            else:
                raise Exception("Public key retrieval failed.")
            
            
    def connect(self):
        """Create a connection with the server

        :param str client: Client source (default = 'api')

        Returns:
        :returns: Connection object (timestamp, connection_id, order_id)
        """

        try:
            headers = {
            'Content-Type': 'text/plain',
            'client': self.client
            }

            aes_key = self.__generate_aes_key()
            self.aes_key = aes_key

            body = {
                "aes_key": aes_key,
                "api_key": self.api_key
            }

            encrypted_body = self.__oaep_encrypt(body)

            res = requests.post(f"{self.server_url}/api/connect/", data=encrypted_body, headers=headers)

            if not res.ok:
                raise Exception({"neuropacsError": f"{res.text}"})

            json = res.json()
            connection_id = json["connectionID"]
            self.connection_id = connection_id
            current_datetime = datetime.now()
            formatted_datetime = current_datetime.strftime("%Y-%m-%d %H:%M:%S")
            return {
                "timestamp": formatted_datetime + " UTC",
                "connection_id": connection_id,
                "aes_key": aes_key,
            }
        except Exception as e:
            if(isinstance(e.args[0], dict) and 'neuropacsError' in e.args[0]):
                raise Exception(e.args[0]['neuropacsError']) 
            else:
                raise Exception("Connection failed.")
            


    def upload_dataset(self, directory, order_id=None, connection_id=None):
        """Upload a dataset to the server

        :param str directory: Path to dataset folder to be uploaded.
        :param str order_id: Base64 order_id
        :param str connection_id: Base64 connection_id

        :return: Upload status code.
        """
        if order_id is None:
            order_id = self.order_id

        if connection_id is None:
            connection_id = self.connection_id

        try:
            self.dataset_upload = True
            
            if isinstance(directory,str):
                if not os.path.isdir(directory):
                    raise Exception({"neuropacsError": "Path not a directory!"}) 
            else:
                raise Exception({"neuropacsError": "Path must be a string!"}) 

            dataset_id = self.__generate_filename()

            total_files = sum(len(filenames) for _, _, filenames in os.walk(directory))

            with tqdm(total=total_files, desc="Uploading", unit="file") as prog_bar:
                for dirpath, _, filenames in os.walk(directory):
                    for filename in filenames:
                        file_path = os.path.join(dirpath, filename)
                        status = self.upload(file_path, dataset_id, order_id, connection_id)
                        if status != 201:
                            raise Exception({"neuropacsError": "Upload failed!"})
                        prog_bar.update(1)  # Update the outer progress bar for each file
            
            return dataset_id

        except Exception as e:
            if(isinstance(e.args[0], dict) and 'neuropacsError' in e.args[0]):
                raise Exception(e.args[0]['neuropacsError']) 
            else:
                raise Exception("Dataset upload failed.")


    def upload(self, data, dataset_id, order_id=None, connection_id=None):
        """Upload a file to the server

        :param str/bytes data: Path of file to be uploaded or byte array
        :param str dataset_id Base64 dataset_id
        :param str order_id: Base64 order_id 
        :param str connection_id: Base64 connection_id

        :return: Upload status code.
        """

        # if not self.dataset_upload:

        if order_id is None:
            order_id = self.order_id

        if connection_id is None:
            connection_id = self.connection_id

        # get file name
        filename = ""
        if isinstance(data,bytes):
            filename = self.__generate_filename()
        elif isinstance(data,str):
            if os.path.isfile(data):
                normalized_path = os.path.normpath(data)
                directories = normalized_path.split(os.sep)
                filename = directories[-1]
            else:
                raise Exception({"neuropacsError": "Path not a file!"})
        else:
            raise Exception({"neuropacsError": "Unsupported data type!"})

        # encrypt order ID
        encrypted_order_id = self.__encrypt_aes_ctr(order_id, "string", "string")

        # create headers
        headers = {"Content-Type": "application/octet-stream",'connection-id': connection_id, 'client': self.client, 'order-id': encrypted_order_id, 'filename': filename, 'dataset-id': dataset_id}

        # get s3 upload params
        res = requests.get(f"{self.server_url}/api/uploadRequest/", headers=headers)

        if not res.ok:
            raise Exception({"neuropacsError": f"{res.text}"})

        decrypted_s3_info = self.__decrypt_aes_ctr(res.text, "json")

        presigned_url = decrypted_s3_info["presignedURL"]

        form = {
            "Content-Disposition": "form-data",
            "filename": filename,
            "name":"test123"
        }

        BOUNDARY = "neuropacs----------"
        DELIM = ";"
        CRLF = "\r\n"
        SEPARATOR="--"+BOUNDARY+CRLF
        END="--"+BOUNDARY+"--"+CRLF
        CONTENT_TYPE = "Content-Type: application/octet-stream"

        header = SEPARATOR
        for key, value in form.items():
            header += f"{key}: {value}"
            header += DELIM
        header += CRLF
        header += CONTENT_TYPE
        header += CRLF + CRLF

        header_bytes = header.encode("utf-8")

        encrypted_order_id = self.__encrypt_aes_ctr(order_id, "string", "string")

        payload_data = None

        if isinstance(data, bytes):
            # encrypted_binary_data = self.__encrypt_aes_ctr(data, "bytes","bytes")
            payload_data = header_bytes + data + END.encode("utf-8")
        
        elif isinstance(data,str):
            with open(data, 'rb') as f:
                binary_data = f.read()
                # encrypted_binary_data = self.__encrypt_aes_ctr(binary_data, "bytes","bytes")
                payload_data = header_bytes + binary_data + END.encode("utf-8")


        res = requests.put(presigned_url, data=payload_data)

        if not res.ok:
            raise Exception({"neuropacsError": f"{res.text}"})

        return 201

    def new_job (self, connection_id=None):
        """Create a new order

        :param str connection_id: Base64 connection_id

        :return: Base64 string order_id.
        """

        if connection_id is None:
            connection_id = self.connection_id

        try:
            headers = {'Content-type': 'text/plain', 'connection-id': connection_id, 'client': self.client}

            res = requests.post(f"{self.server_url}/api/newJob/", headers=headers)

            if not res.ok:
                raise Exception({"neuropacsError": f"{res.text}"})

            text = res.text
            decrypted_text = self.__decrypt_aes_ctr(text, "string")
            self.order_id = decrypted_text
            return decrypted_text
        except Exception as e:
            if(isinstance(e.args[0], dict) and 'neuropacsError' in e.args[0]):
                raise Exception(e.args[0]['neuropacsError'])
            else:
                raise Exception("Job creation failed.")            


    def run_job(self, product_id, order_id=None, dataset_id=None, connection_id=None):
        """Run a job
        
        :param str productID: Product to be executed.
        :prarm str order_id: Base64 order_id 
        :prarm str dataset_id: Base64 dataset_id 
        :param str connection_id: Base64 connection_id
        
        :return: Job run status code.
        """
        if order_id == None:
            order_id = self.order_id

        if connection_id is None:
            connection_id = self.connection_id

        try:
            headers = {'Content-type': 'text/plain', 'connection-id': connection_id, 'client': self.client}

            body={}
            if dataset_id is None:
                body = {
                    'orderID': order_id,
                    'productID': product_id,
                }
            else:
                body = {
                    'orderID': order_id,
                    'productID': product_id,
                    'datasetID': dataset_id
                }

            encryptedBody = self.__encrypt_aes_ctr(body, "json", "string")

            res = requests.post(f"{self.server_url}/api/runJob/", data=encryptedBody, headers=headers)
            
            if not res.ok:
                raise Exception({"neuropacsError": f"{res.text}"})

            return res.status_code
                
        except Exception as e:
            if(isinstance(e.args[0], dict) and 'neuropacsError' in e.args[0]):
                raise Exception(e.args[0]['neuropacsError'])
            else:
                raise Exception("Job run failed.")  


    def check_status(self, order_id=None, dataset_id=None, connection_id=None):
        """Check job status

        :param str order_id: Base64 order_id (optional)
        :param str dataset_id: Base64 dataset_id (optional)
        :param str connection_id: Base64 connection_id

        :return: Job status message.
        """
        if order_id is None:
            order_id = self.order_id

        if connection_id is None:
            connection_id = self.connection_id

        try:

            headers = {'Content-type': 'text/plain', 'connection-id': connection_id, 'client': self.client}

            if dataset_id is not None:
                body = {
                    'orderID': order_id,
                    'datasetID': dataset_id
                }
            else:
               body = {
                    'orderID': order_id,
                }

            encryptedBody = self.__encrypt_aes_ctr(body, "json", "string")

            res = requests.post(f"{self.server_url}/api/checkStatus/", data=encryptedBody, headers=headers)
            
            if not res.ok:
                raise Exception({"neuropacsError": f"{res.text}"})
            
            text = res.text
            json = self.__decrypt_aes_ctr(text, "json")
            return json
            
        except Exception as e:
            if(isinstance(e.args[0], dict) and 'neuropacsError' in e.args[0]):
                raise Exception(e.args[0]['neuropacsError'])
            else:
                raise Exception("Status check failed.")  


    def get_results(self, format, order_id=None, dataset_id=None, connection_id=None):
        """Get job results

        :param str format: Format of file data
        :prarm str order_id: Base64 order_id (optional)
        :param str connection_id: Base64 connection_id

        :return: AES encrypted file data in specified format
        """

        if order_id is None:
            order_id = self.order_id

        if connection_id is None:
            connection_id = self.connection_id
        
        try:
        
            headers = {'Content-type': 'text/plain', 'connection-id': connection_id, 'client': self.client}

            if dataset_id is not None:
                body = {
                    'orderID': order_id,
                    'format': format,
                    'datasetID': dataset_id
                }
            else:
                body = {
                    'orderID': order_id,
                    'format': format               
                }

            validFormats = ["TXT", "XML", "JSON", "PDF", "DCM"]

            if format not in validFormats:
                raise Exception({"neuropacsError" : "Invalid format! Valid formats include: \"TXT\", \"JSON\", \"XML\", \"PDF\", \"DCM\" ."})

            encrypted_body = self.__encrypt_aes_ctr(body, "json", "string")

            res = requests.post(f"{self.server_url}/api/getResults/", data=encrypted_body, headers=headers)
            
            if not res.ok:
                raise Exception({"neuropacsError": f"{res.text}"})

            text = res.text
            decrypted_file_data = self.__decrypt_aes_ctr(text, "string")
            return decrypted_file_data

        except Exception as e:
            if(isinstance(e.args[0], dict) and 'neuropacsError' in e.args[0]):
                raise Exception(e.args[0]['neuropacsError'])
            else:
                raise Exception(f"Result retrieval failed!")


    


import unittest
import re
import json
import time
from neuropacs.sdk import Neuropacs
from datetime import datetime, timezone
from cryptography.hazmat.primitives import serialization


# RUN TESTS: python -m unittest tests.tests_neuropacs

'''
Test Scenarios

- invalid server url
- get public key
    - success X
- connect 
    - success X
    - invalid api key X
- new job 
    - success X
    - invalid connection id X
- upload 
    - File object success X
    - Uint8Array success X
- upload dataset 
    - dataset path success X
- run job
    - success X
    - invalid product X
    - invalid order id X
    - invalid connection id X
- check status
    - success X
    - invalid order id X
    - invalid connection id X
- get results
    - success X
        - TXT X
        - JSON X
        - XML X
    - invalid result format X
    - invalid order id X
    - invalid connection id X
'''

# SET THIS TO FALSE BEFORE PUSHING TO GITHUB
test_env = False

test_log_path = ''

test_results = {}

start_time = 0
end_time = 0

def log_test(description, input_data, expected_output, actual_output):
    if test_env:
        log_message = """
    Test Description: {description}

        Input:
        {input_data}

        Expected Output:
        {expected_output}

        Actual Output:
        {actual_output}
    ----------------------------------------------------------------------""".format(
            description=description,
            input_data=str(input_data),
            expected_output=str(expected_output),
            actual_output=str(actual_output)
        )

        with open(test_log_path, 'a') as file:
            file.write(log_message)
            file.close()

def log_result(status, fail_desc=None):
    if test_env: 
        status_message = f"""
    Test result: {status}\n"""
        if fail_desc is not None:
            status_message += f"""
    Description: {fail_desc}\n\n"""
        else:
            status_message += "\n\n"

        with open(test_log_path, 'a') as file:
            file.write(status_message)
            file.close()

def log_test_results():
    if test_env:
        total_tests = 0
        passed_tests = 0

        with open(test_log_path, 'a') as file:
            for test in test_results:
                if test_results[test] == "PASS":
                    result_str = f"O - {test}"
                    passed_tests += 1
                    total_tests += 1
                elif test_results[test] == "FAIL":
                    result_str = f"X - {test}"
                    total_tests += 1
                file.write(result_str + "\n")
            test_score_str = f"""
    Test Suites: {1} passed, {1} total
    Tests: {passed_tests} passed, {total_tests} total
    Ran all test suites in {end_time-start_time}s.
    """
            file.write(test_score_str)
            file.close()


def is_valid_public_key(base64_pem_key):
    """
    Validate a Base64 PEM public key.

    Args:
        public_key (str): Base64 PEM Public key.

    Returns:
        bool: Status of the public key. True if the key is valid, False otherwise.
    """
    try:
        '''
        type: "spki",
        format: "pem"
        '''
        encoded_key = base64_pem_key.encode('utf-8')
        public_key = serialization.load_pem_public_key(encoded_key)
        return True
    except Exception as error:
        print(error)
        return False  # If an error occurs, the key is not valid

class UnitTests(unittest.TestCase):

    def setUp(self):
        server_url = "http://ec2-3-142-212-32.us-east-2.compute.amazonaws.com:5000"
        api_key = "m0ig54amrl87awtwlizcuji2bxacjm"
        self.npcs = Neuropacs(server_url, api_key)          
        
    @classmethod
    def setUpClass(cls):
        if test_env:
            global start_time
            start_time = time.time()
            current_date = datetime.utcnow()
            formatted_utc_string = current_date.strftime("%m-%d-%yT%H-%M-%S")
            file_path = f"./tests/test_logs/unittest_{formatted_utc_string}.log"
            global test_log_path
            test_log_path = file_path
            titleMessage = f"Python SDK Unit Tests\nTest run date: {formatted_utc_string}\n\n"
            with open(test_log_path, 'a') as file:
                file.write(titleMessage)
                file.close()

    @classmethod
    def tearDownClass(cls):
        if test_env:
            global end_time
            end_time = time.time()
            log_test_results()
        
    # Test 1: get_public_key() - Success
    def test_get_public_key(self):

        public_key = self.npcs.get_public_key() #Get public key

        ### TEST PUBLIC KEY ###
        try:
            validPubKey = is_valid_public_key(public_key)
            self.assertTrue(validPubKey)
            self.assertIsInstance(public_key, str, "Public key should be a string")
            public_key_regex = re.compile(r'-----BEGIN PUBLIC KEY-----\r?\n([A-Za-z0-9+/=\r\n]+)\r?\n-----END PUBLIC KEY-----')
            self.assertTrue(public_key_regex.match(public_key), "Public key should be a valid base64 encoded key")
            log_test("Successfully retrieve public key.", None, "Valid PEM formatted base64 encoded public key", public_key)
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_get_public_key'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_get_public_key'] = "PASS"

        ### TEST PUBLIC KEY ###

    # Test 2: connect() - SUCCESS
    def test_connect_success(self):
    
        connection = self.npcs.connect() # Get connection

        connection_id = connection['connection_id']
        aes_key = connection['aes_key']

        ### TEST CONNECTION ID ###
        try:
            self.assertIsInstance(connection_id, str, "Connection ID should be a string")
            self.assertEqual(len(connection_id), 32, "Connection ID should be length 32")
            self.assertTrue(connection_id.isalnum(), "Connection ID should contain only alphanumeric characters")
            self.assertIsInstance(aes_key, str, "Key should be a string")
            self.assertEqual(len(aes_key), 24, "Key should be length 24")
            aes_key_regex = re.compile(r'^[A-Za-z0-9+/]{22}==$')
            self.assertTrue(aes_key_regex.match(aes_key), "Key should be a valid base64 encoded key")
            ### TEST CONNECTION ID ###
            log_test("Successfully create a connection.",{"aes_key": self.npcs.aes_key,"api_key": self.npcs.api_key},"Connection object containing AES key, connection ID, and timestamp", connection)
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_connect_success'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_connect_success'] = "PASS"

    # Test 3: connect() - INVALID API KEY
    def test_connect_invalid_api_key_fail(self):
        
        invalid_api_key = "thisisnotarealapikey12345"
        self.npcs.api_key = invalid_api_key

        try:
            with self.assertRaises(Exception) as context:
                connection = self.npcs.connect() # Get connection
            ### TEST CONNECTION ID FAIL ###
            self.assertEqual(str(context.exception),"Connection failed!")  
            ### TEST CONNECTION ID FAIL ###
            log_test("Fail while creating a connection due to invalid API key.", {"aes_key": self.npcs.aes_key,"api_key": self.npcs.api_key}, "Connection failed!", str(context.exception))
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_connect_invalid_api_key_fail'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_connect_invalid_api_key_fail'] = "PASS"

        
    # Test 4: new_job() - SUCCESS
    def test_new_job_success(self):
        
        self.npcs.connect() # Get connection

        order_id = self.npcs.new_job() # Create new order
        try:
            ### TEST ORDER ID ###        
            self.assertIsInstance(order_id, str, "Order ID should be a string")
            self.assertEqual(len(order_id), 20, "Order ID should be length 32")
            self.assertTrue(order_id.isalnum(), "Order ID should contain only alphanumeric characters")
            ### TEST ORDER ID ###
            log_test("Successfully create a new job and return an order ID.", {'connection-id': self.npcs.connection_id}, "Order ID (string, length 20, alphanumeric)", order_id)
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_new_job_success'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_new_job_success'] = "PASS"

    # Test 5: new_job() - INVALID CONNECTION ID
    def test_new_job_invalid_connection_id_fail(self):
        
        invalid_connection_id = "thisisnotarealconnectionid12345"
        self.npcs.connection_id = invalid_connection_id
        try:
            with self.assertRaises(Exception) as context:
                order_id = self.npcs.new_job() # Create new order
            ### TEST ORDER ID FAIL ###  
            self.assertEqual(str(context.exception), "Job creation failed!") 
            ### TEST ORDER ID FAIL ###
            log_test("Fail while creating a new job due to invalid connection ID.", {'connection-id': self.npcs.connection_id}, "Job creation failed!", str(context.exception))
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_new_job_invalid_connection_id_fail'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_new_job_invalid_connection_id_fail'] = "PASS"


    # Test 6: upload() - SUCCESS (path)
    def test_upload_by_path_success(self):

        test_dicom_file_path = "./tests/test_dataset/testdcm"

        self.npcs.connect() # get connection

        self.npcs.new_job() # Create new order

        upload_status = self.npcs.upload(test_dicom_file_path) # Upload file

        try:
            ### TEST UPLOAD STATUS ###
            self.assertEqual(upload_status, 201)
            ### TEST UPLOAD STATUS ###
            log_test("Successfully upload a file via path string.", test_dicom_file_path, 201, upload_status)
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_upload_by_path_success'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_upload_by_path_success'] = "PASS"

    # Test 7: upload() - SUCCESS (bytes)
    def test_upload_by_bytes_success(self):

        with open('./tests/test_dataset/testdcm', 'rb') as file:
            test_dicom_bytes = file.read()

        self.npcs.connect() # Get connection

        self.npcs.new_job() # Create new order

        upload_status = self.npcs.upload(test_dicom_bytes) # Upload file

        try:
            ### TEST UPLOAD STATUS ###
            self.assertEqual(upload_status, 201)
            ### TEST UPLOAD STATUS ###
            log_test("Successfully upload a file via byte array (bytes).", test_dicom_bytes[:50], 201, upload_status)
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_upload_by_bytes_success'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_upload_by_bytes_success'] = "PASS"


    # Test 8: upload_dataset() - SUCCESS (path)
    def test_upload_dataset_by_path_success(self):

        test_dicom_file_path = "./tests/test_dataset/"

        self.npcs.connect() # Get connection

        self.npcs.new_job() # Create new order

        upload_status = self.npcs.upload_dataset(test_dicom_file_path) # Upload file

        try:
            ### TEST UPLOAD STATUS ###
            self.assertEqual(upload_status, 201)
            ### TEST UPLOAD STATUS ###
            log_test("Successfully upload a dataset via path string.", test_dicom_file_path, 201, upload_status)
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_upload_dataset_by_path_success'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_upload_dataset_by_path_success'] = "PASS"

    # Test 9: run_job() - SUCCESS
    def test_run_job_success(self):

        product_id = "PD/MSA/PSP-v1.0" # product ID

        self.npcs.connect() # Get connection id

        self.npcs.new_job() # Create new order

        job = self.npcs.run_job(product_id) # Run job

        try:
            ### TEST UPLOAD STATUS ###
            self.assertEqual(job, 202)
            ### TEST UPLOAD STATUS ###
            log_test("Successfully run a job and return a status 202.", {'orderID': self.npcs.order_id,'productID': product_id}, 202, job)
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_run_job_success'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_run_job_success'] = "PASS"
        

    # Test 10: run_job() - INVALID PRODUCT ID
    def test_run_job_invalid_product_id_fail(self):

        product_id = "notARealProduct" # product ID

        self.npcs.connect() # Get connection 

        self.npcs.new_job() # Create new order

        try:
            with self.assertRaises(Exception) as context:
                job = self.npcs.run_job(product_id) # Run job

            ### TEST JOB RUN FAIL ###  
            self.assertEqual(str(context.exception), "Job run failed.") 
            ### TEST JOB RUN FAIL ###
            log_test("Fail running a job due to invalid product ID.", {'connection-id':self.npcs.connection_id, 'orderID': self.npcs.order_id,'productID': product_id}, "Job run failed.", str(context.exception))
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_run_job_invalid_product_id_fail'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_run_job_invalid_product_id_fail'] = "PASS"


    # Test 11: run_job() - INVALID ORDER ID
    def test_run_job_invalid_order_id_fail(self):

        product_id = "PD/MSA/PSP-v1.0" # product ID

        self.npcs.connect() # Get connection

        try:
            with self.assertRaises(Exception) as context:
                job = self.npcs.run_job(product_id, "notARealOrderID") # Run job

            ### TEST JOB RUN FAIL ###  
            self.assertEqual(str(context.exception), "Job run failed.") 
            ### TEST JOB RUN FAIL ###
            log_test("Fail running a job due to invalid order ID.", {'connection-id':self.npcs.connection_id, 'orderID': "notARealOrderID" ,'productID': product_id}, "Job run failed.", str(context.exception))
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_run_job_invalid_order_id_fail'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_run_job_invalid_order_id_fail'] = "PASS"


    # Test 12: run_job() - INVALID CONNECTION ID
    def test_run_job_invalid_connection_id_fail(self):

        product_id = "PD/MSA/PSP-v1.0" # product ID

        self.npcs.connect() # Get connection id

        self.npcs.new_job() # Create new order

        self.npcs.connection_id = "notARealConnectionID"

        try:
            with self.assertRaises(Exception) as context:
                job = self.npcs.run_job(product_id) # Run job

            ### TEST JOB RUN FAIL ###  
            self.assertEqual(str(context.exception), "Job run failed.") 
            ### TEST JOB RUN FAIL ###
            log_test("Fail running a job due to invalid connection ID.", {'connection-id':self.npcs.connection_id, 'orderID': self.npcs.order_id,'productID': product_id}, "Job run failed.", str(context.exception))
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_run_job_invalid_connection_id_fail'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_run_job_invalid_connection_id_fail'] = "PASS"

    # Test 13: check_status() - SUCCESS
    def test_check_status_success(self):

        product_id = "PD/MSA/PSP-v1.0" # product ID

        self.npcs.connect() # Get connection

        self.npcs.new_job() # Create new order

        self.npcs.run_job(product_id) # Run job

        job_status = self.npcs.check_status("TEST") # Check status

        try:
            ### TEST CHECK STATUS ###
            self.assertIsInstance(job_status, str, "Job status should be a string")
            self.assertEqual(job_status, r'{"started":true,"finished":true,"failed":false,"progress":100,"info":"Finished"}')
            ### TEST CHECK STATUS ###
            log_test("Successfully check status of job.", {'connection-id': self.npcs.connection_id, 'orderID': "TEST"},r'{"started":true,"finished":true,"failed":false,"progress":100,"info":"Finished"}', job_status)
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_check_status_success'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_check_status_success'] = "PASS"


    # Test 14: check_status() - INVALID ORDER ID
    def test_check_status_invalid_order_id_fail(self):

        self.npcs.connect() # Get connection

        try:
            with self.assertRaises(Exception) as context:
                job_status = self.npcs.check_status("invalidOrderID") # check status

            ### TEST CHECK STATUS FAIL ###  
            self.assertEqual(str(context.exception), "Status check failed.") 
            ### TEST CHECK STATUS FAIL ###
            log_test("Fail checking job status due to invalid order ID.", {'connection-id': self.npcs.connection_id, 'orderID': "invalidOrderID"}, "Status check failed.", str(context.exception))
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_check_status_invalid_order_id_fail'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_check_status_invalid_order_id_fail'] = "PASS"

   # Test 15: check_status() - INVALID CONNECTION ID
    def test_check_status_invalid_connection_id_fail(self):

        self.npcs.connect() # Get connection

        self.npcs.connection_id = "noARealConnectionID"

        try:
            with self.assertRaises(Exception) as context:
                job_status = self.npcs.check_status("TEST") # check status

            ### TEST CHECK STATUS FAIL ###  
            self.assertEqual(str(context.exception), "Status check failed.") 
            ### TEST CHECK STATUS FAIL ###
            log_test("Fail checking job status due to invalid connection ID.", {'connection-id':"noARealConnectionID", 'orderID': "TEST"}, "Status check failed.", str(context.exception))
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_check_status_invalid_connection_id_fail'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_check_status_invalid_connection_id_fail'] = "PASS"

    # Test 16: get_results() - SUCCESS (TXT)
    def test_get_results_txt_success(self):

        self.npcs.connect() # Get connection

        job_results = self.npcs.get_results("TXT", "TEST") # Get results

        current_date = datetime.now(timezone.utc).strftime("%Y-%m-%d")

        try:
            ### TEST JOB RESULTS ###
            self.assertIsInstance(job_results, str, "Job results should be a string")
            self.assertEqual(job_results, f"Order ID: TEST\nDate: {current_date}\nProduct: TEST\nPD probability vs. MSA/PSP: 62.6%\nMSA probability vs. PSP: 85.6%\nBiomarker levels: pSN=0.26, Putamen=0.19, SCP=0.48, MCP=0.07")
            ### TEST JOB RESULTS ###
            log_test("Successfully retrieve job results in TXT format.", {'connection-id': self.npcs.connection_id, 'orderID': "TEST", 'format': "TXT"}, f"Order ID: TEST\nDate: {current_date}\nProduct: TEST\nPD probability vs. MSA/PSP: 62.6%\nMSA probability vs. PSP: 85.6%\nBiomarker levels: pSN=0.26, Putamen=0.19, SCP=0.48, MCP=0.07", job_results)
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_get_results_txt_success'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_get_results_txt_success'] = "PASS"

    # Test 17: get_results() - SUCCESS (JSON)
    def test_get_results_json_success(self):

        self.npcs.connect() # Get connection

        job_results = self.npcs.get_results("JSON", "TEST") # Get results

        current_date = datetime.now(timezone.utc).strftime("%Y-%m-%d")
        expected_result_pt1 = r'{"orderID":"TEST","date":"'
        expected_result_pt2 = r'","product":"TEST","result":{"PDprobability":"62.6","MSAprobability":"85.6","FWpSN":"0.26","FWPutamen":"0.19","FWSCP":"0.48","FWMCP":"0.07"}}'
        expected_result_full = expected_result_pt1 + current_date + expected_result_pt2

        try:
            ### TEST JOB RESULTS ###
            self.assertEqual(re.sub(r'\s', '', job_results), expected_result_full)
            try:
                json_object = json.loads(job_results)
                self.assertIsInstance(json_object, dict)
            except json.JSONDecodeError:
                self.fail("Job results must be valid JSON.")
            ### TEST JOB RESULTS ###
            log_test("Successfully retrieve job results in JSON format.", {'connection-id': self.npcs.connection_id, 'orderID': "TEST", 'format': "JSON"}, expected_result_full, re.sub(r'\s', '', job_results))
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_get_results_json_success'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_get_results_json_success'] = "PASS"


    # Test 18: get_results() - SUCCESS (XML)
    def test_get_results_xml_success(self):

        self.npcs.connect() # Get connection

        job_results = self.npcs.get_results("XML", "TEST") # Get results

        expected_result_pt1 = r'<?xmlversion="1.0"encoding="UTF-8"standalone="yes"?><neuropacsorderID="TEST"date="'
        current_date = datetime.now(timezone.utc).strftime("%Y-%m-%d")
        expected_result_pt2 = r'"product="TEST"><resultname="PDprobability"value="62.6"/><resultname="MSAprobability"value="85.6"/><dataname="FWpSN"value="0.26"/><dataname="FWPutamen"value="0.19"/><dataname="FWSCP"value="0.48"/><dataname="FWMCP"value="0.07"/></neuropacs>'
        expected_result_full = expected_result_pt1 + current_date + expected_result_pt2

        try:
            ### TEST JOB RESULTS ###
            self.assertEqual(re.sub(r'\s', '', job_results), expected_result_full)
            self.assertIsInstance(job_results, str, "Job results should be a string XML")
            ### TEST JOB RESULTS ###
            log_test("Successfully retrieve job results in XML format.", {'connection-id': self.npcs.connection_id, 'orderID': "TEST", 'format': "XML"}, expected_result_full, re.sub(r'\s', '', job_results))
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_get_results_xml_success'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_get_results_xml_success'] = "PASS"


   # Test 19: get_status() - INVALID RESULT FORMAT
    def test_check_status_invalid_result_format_fail(self):

        self.npcs.connect() # Get connection id

        try:
            with self.assertRaises(Exception) as context:
                self.npcs.get_results("noARealFormat", "TEST") # Get results

            ### TEST JOB RESULTS FAIL ###  
            self.assertEqual(str(context.exception), r'Invalid format! Valid formats include: "TXT", "JSON", "XML".') 
            ### TEST JOB RESULTS FAIL ###
            log_test("Fail to check job status due to invalid result format.", {'connection-id': self.npcs.connection_id, 'orderID': "TEST", 'format': "noARealFormat"},r'Invalid format! Valid formats include: "TXT", "JSON", "XML".', str(context.exception))
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_check_status_invalid_result_format_fail'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_check_status_invalid_result_format_fail'] = "PASS"

   # Test 20: get_status() - INVALID ORDER ID
    def test_check_status_invalid_order_id_fail(self):

        self.npcs.connect() # Get connection 

        try:
            with self.assertRaises(Exception) as context:
                self.npcs.get_results("TXT", "noARealOrderID") # Get results

            ### TEST JOB RESULTS FAIL ###  
            self.assertEqual(str(context.exception), "Result retrieval failed!") 
            ### TEST JOB RESULTS FAIL ###
            log_test("Fail to check job status due to invalid order ID.", {'connection-id': self.npcs.connection_id, 'orderID': "noARealOrderID", 'format': "TXT"},"Result retrieval failed!", str(context.exception))
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_check_status_invalid_order_id_fail'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_check_status_invalid_order_id_fail'] = "PASS"


   # Test 21: get_status() - INVALID CONNECTION ID
    def test_check_status_invalid_connection_id_fail(self):

        self.npcs.connect() # Get connection

        self.npcs.connection_id = "noARealConnectionID"

        try:
            with self.assertRaises(Exception) as context:
                self.npcs.get_results("TXT", "TEST") # Get results

            ### TEST JOB RESULTS FAIL ###  
            self.assertEqual(str(context.exception), "Result retrieval failed!") 
            ### TEST JOB RESULTS FAIL ###
            log_test("Fail to check job status due to invalid connection ID.", {'connection-id': self.npcs.connection_id, 'orderID': "TEST", 'format': "TXT"},"Result retrieval failed!", str(context.exception))
        except AssertionError as e:
            log_result("FAIL", e)
            test_results['test_check_status_invalid_connection_id_fail'] = "FAIL"
        else:
            log_result("PASS")
            test_results['test_check_status_invalid_connection_id_fail'] = "PASS"










